"""Interface for standard module logging"""

# Standard modules
import logging
import os
import sys

# Local modules
from . import file


def setup_log_directory(path: str, file_limit: int = 10) -> None:
    """Make the log directory and delete old log files

    path: the log directory path
    file_limit: how many old log files to keep
    """

    # Make the directory
    file.make_dir(path)

    # Delete old log files
    log_files = []
    for entry in os.listdir(path):
        if file.is_file(file.join(path, entry)):
            if file.file_type(entry) == "log":
                log_files.append(entry)
    if len(log_files) > file_limit:
        for i in range(file_limit, len(log_files)):
            file.delete(file.join(path, log_files[i - file_limit]))


def setup_log_file(path: str, head: str, print_head: bool = True) -> None:
    """Create the log file and write the head (additionally print the head to console)

    path: the log file path
    head: a head of the application (for example for a description, version, author ...)
    print_head: print the head to console
    """

    # Print the head
    if print_head:
        print(head)

    # Create and open the file to write the head
    with file.open_utf8(path, "w") as f:
        f.write(head + "\n")


def create_logger(path: str, name: str, level: int, *, log_date: bool = True, log_thread: bool = False, log_custom: str = "", log_custom_date: str = "") -> logging.Logger:
    """Create the logger

    path: the log file path
    name: the name of the logger
    level: the minimum level of the logs to show
    log_date: show the date (if no custom format)
    log_thread: show the thread name (if no custom format)
    log_custom: custom format (default format if empty string)
    log_custom_date: custom date format (default date format if empty string)
    """

    # Define format
    log_format = logging.Formatter(f"[%(asctime)s] [%(name)s{' - %(threadName)s' if log_thread else ''}] [%(levelname)s] %(message)s" if log_custom == "" else log_custom,
                                   datefmt=('%d/%b/%y %H:%M:%S' if log_date else '%H:%M:%S') if log_custom_date == "" else log_custom_date)

    # Setup handler
    log_handler_file = logging.FileHandler(path, encoding="utf-8")
    log_handler_console = logging.StreamHandler(sys.stdout)
    log_handler_file.setFormatter(log_format)
    log_handler_console.setFormatter(log_format)

    # Create logger
    logger = logging.Logger(name, level)
    logger.addHandler(log_handler_file)
    logger.addHandler(log_handler_console)

    # Return logger
    return logger



"""Interface for standard modules os and json"""

# Standard modules
import json
import os
from logging import Logger
from typing import IO


def open_file(path: str, mode: str, logger: Logger = None, *, create: bool = True, encoding: str = None) -> IO:
    """Open a file

    path: the path of the file
    mode: the mode to use to open the file
    logger: the logger to log information and warnings
    create: do create directories and file, if not existent
    encoding: the encoding to use for the file
    """

    # Check for directory
    if create and not exists(directory(path)) and not directory(path) == "":
        if logger:
            logger.info(f"Doesn't found directory '{directory(path)}'! Create directory ...")
        os.makedirs(directory(path))

    try:

        # Open file
        if logger:
            logger.debug(f"Open file '{path}' with mode '{mode}' ...")
        return open(path, mode, encoding=encoding)

    except OSError:

        # Try to handle error
        if create:
            if logger:
                logger.info(f"Failed to open file '{path}'! Create and try to load it again ...")
            open(path, "w", encoding=encoding).close()
            return open_file(path, mode, logger, create=False)
        else:
            if logger:
                logger.warning(f"Failed to open file '{path}'! Raise error ...")
            raise


def open_utf8(path: str, mode: str, logger: Logger = None, *, create: bool = True) -> IO:
    """Open a file with UTF-8 encoding

    path: the path of the file
    mode: the mode to use to open the file
    logger: the logger to log information and warnings
    create: do create directories and file, if not existent
    """

    return open_file(path, mode, logger, create=create, encoding="utf-8")


def load_json(path: str, logger: Logger = None, *, create: bool = True) -> dict:
    """Load a json file to a dictionary

    path: the path of the file
    logger: the logger to log information and warnings
    create: do create directories and file, if not existent
    """

    # Return empty dictionary on missing file
    if not create and not exists(path):
        logger.warning(f"Doesn't found file '{path}'! Continue with empty dictionary ...")
        return {}

    try:

        # Open and load dictionary
        with open_utf8(path, "r", logger) as f:
            return json.load(f)

    except json.JSONDecodeError:

        # Return empty dictionary of error
        logger.warning(f"Failed to decode file '{path}' to json! Continue with empty dictionary ...")
        return {}


def save_json(path: str, data: dict, logger: Logger = None) -> None:
    """Load a json file to a dictionary

    path: the path of the file
    data: the dictionary to save
    logger: the logger to log information and warnings
    """

    with open_utf8(path, "w", logger) as f:
        json.dump(data, f, indent=4, separators=(', ', ': '))


def delete(path: str, logger: Logger = None, *, ignore_error: bool = False) -> None:
    """Delete a file"""

    if logger:
        logger.debug(f"Delete file '{path}' ...")

    try:
        os.remove(path)
    except OSError:
        if not ignore_error:
            if logger:
                logger.warning(f"Failed to delete file '{path}'! Raise error ...")
            raise
        if logger:
            logger.info(f"Failed to delete file '{path}'!")


def make_dir(path: str, logger: Logger = None, *, parent: bool = True, ignore_error: bool = False) -> None:
    """Make a directory"""

    if logger:
        logger.debug(f"Make directory at '{path}' ...")

    try:
        if parent:
            os.makedirs(path, exist_ok=True)
        else:
            os.mkdir(path)
    except OSError:
        if not ignore_error:
            if logger:
                logger.warning(f"Failed to make directory at '{path}'! Raise error ...")
            raise
        if logger:
            logger.info(f"Failed to make directory at '{path}'!")


def remove_dir(path: str, logger: Logger = None, *, parent: bool = False, ignore_error: bool = False) -> None:
    """Remove a directory"""

    if logger:
        logger.debug(f"Delete directory at '{path}' ...")

    try:
        if parent:
            os.removedirs(path)
        else:
            os.rmdir(path)
    except OSError:
        if not ignore_error:
            if logger:
                logger.warning(f"Failed to remove directory at '{path}'! Raise error ...")
            raise
        if logger:
            logger.info(f"Failed to remove directory at '{path}'!")


def exists(path: str) -> bool:
    """Check file existence"""

    return os.path.exists(path)


def name(path: str) -> str:
    """Get the basename of a path"""

    return os.path.basename(path)


def directory(path: str) -> str:
    """Get the directory of a path"""

    return os.path.dirname(path)


def file_type(path: str) -> str:
    """Get the type of a file"""

    return name(path).split(".")[-1]


def file_name(path: str) -> str:
    """Get the name of a file without file extension"""

    return ".".join(name(path).split(".")[0:-2]) if len(name(path)) > 1 else ""


def is_file(path: str) -> bool:
    """Is file"""

    return os.path.isfile(path)


def is_dir(path: str) -> bool:
    """Is directory"""

    return os.path.isdir(path)


def join(dir_path: str, filename: str) -> str:
    """Join a directory path with a filename"""

    return os.path.join(dir_path, filename)


def rename(old_path: str, new_path: str, logger: Logger = None) -> None:
    """Rename a file or path"""

    if logger:
        logger.debug(f"Rename '{old_path}' to '{new_path}' ...")

    os.rename(old_path, new_path)


def dir_size(path: str, logger: Logger = None) -> int:
    """Get the size of a directory"""

    size = 0

    try:

        if logger:
            logger.debug(f"Get the size of '{path}' ...")

        for entry in os.scandir(path):

            if entry.is_file():
                size += file_size(entry.path)
            elif entry.is_dir():
                size += dir_size(entry.path, logger)

    except PermissionError:
        if logger:
            logger.info(f"Failed to get the size of '{path}'! Unresolved permission! Continue with 0 ...")
        return 0
    except OSError:
        if logger:
            logger.warning(f"Failed to get the size of '{path}'! Unexpected error! Continue with 0 ...")
        return 0

    return size


def file_size(path: str) -> int:
    """Get the size of a file"""

    return os.path.getsize(path)


def format_size_1000(size: int) -> str:
    """Format a size with a factor of 1000"""

    for i in ["", "K", "M", "G", "T", "P"]:

        if size < 1000:
            return f"{size:.2f}{i}B"

        size /= 1000

    return f"{size:.2f}EB"


def format_size_1024(size: int) -> str:
    """Format a size with a factor of 1024"""

    for i in ["", "Ki", "Mi", "Gi", "Ti", "Pi"]:

        if size < 1024:
            return f"{size:.2f}{i}B"

        size /= 1024

    return f"{size:.2f}EiB"



# External modules
import pygame as pg


class FontManager:
    """Font manager class"""
    
    def __init__(self) -> None:

        self.fonts = {}
        self.cache = {}

    def define(self, name: str, path: str) -> None:
        """Define a new font"""
        self.fonts[name] = path

    def get(self, name: str, size: int, system: bool = False) -> pg.font.Font:
        """Get a font"""

        # Check for cache
        if name in self.cache:
            if size in self.cache[name]:
                return self.cache[name][size]

        # Load font
        if system:
            font = pg.font.SysFont(name, size)
        else:
            font = pg.font.Font(self.fonts[name], size)

        # Store to cache
        if name not in self.cache:
            self.cache[name] = {}
        self.cache[name][size] = font

        # Return cache
        return font



# Standard modules
from __future__ import annotations
from dataclasses import dataclass, field
from typing import Callable, TYPE_CHECKING

# External modules
import pygame as pg

# Local modules
if TYPE_CHECKING:
    from .window import Window


@dataclass(frozen=True)
class EventHook:
    """Event hook dataclass"""

    name: str
    events: tuple[int, ...]
    handlers: tuple[Callable[[pg.event.Event, Window, dict], None], ...]
    data: dict



"""Contains the enum class and metaclass"""


class EnumError(Exception):
    """A error for invalid usage of enums"""
    pass


class EnumMeta(type):
    """The metaclass for an enum"""

    __unique = 1

    def __new__(mcs, name, bases, attrs):

        # Create class
        attrs["__build__"] = None
        __enum = super().__new__(mcs, name, bases, attrs)

        # Generate enum elements
        for attr, value in __enum.__annotations__.items():
            if value == int:
                setattr(__enum, attr, mcs.__unique)
                mcs.__unique += 1
            elif value == str:
                setattr(__enum, attr, attr)

        # Return class
        delattr(__enum, "__build__")
        return __enum

    # OVERLOADS

    def __setattr__(cls, key, value):
        if "__build__" not in vars(cls):
            raise EnumError("Changing enum elements is not allowed!")
        super().__setattr__(key, value)

    def __delattr__(cls, key):
        if "__build__" not in vars(cls):
            raise EnumError("Deleting enum elements is not allowed!")
        super().__delattr__(key)

    def __contains__(cls, item):
        try:
            return item in getattr(cls, "values")
        except (IndexError, TypeError, KeyError):
            return False

    def __iter__(cls):
        return iter(getattr(cls, "values"))

    def __reversed__(cls):
        return reversed(getattr(cls, "values"))

    def __len__(cls):
        return len(getattr(cls, "values"))

    def __getitem__(cls, item):
        if item not in cls:
            raise EnumError(f"Cannot find '{item}' in '{cls}'!")
        return getattr(cls, item)

    def __setitem__(cls, item, value):
        raise EnumError("Changing enum elements is not allowed!")

    def __delitem__(cls, item):
        raise EnumError("Deleting enum elements is not allowed!")

    def __call__(cls, element: str):
        if element not in cls:
            raise EnumError(f"Cannot find '{element}' in '{cls}'!")
        return getattr(cls, element)

    # PROPERTIES

    @property
    def values(cls) -> list:
        """Returns all enum elements as a list"""
        return [value for name, value in vars(cls).items() if not name.startswith("__") and name != "values"]


class Enum(metaclass=EnumMeta):
    """The enum class

    Integer enum:
    elem1: int
    elem2: int
    ...

    Generates a enum with automatic assignment of a unique number


    String enum:
    elem1: str
    elem2: str
    ...

    Generates a enum with automatic assignment of the element name as string


    Custom enum:
    elem1 = Class(arg1, arg2, ...)
    elem2 = Class(arg1, arg2, ...)
    ...

    Generates a enum with custom objects

    """



# Standard modules
from abc import ABC, abstractmethod
from typing import Self

# External modules
import pygame as pg

# Local modules
from .vector2d import Vec


class Renderer(ABC):
    """Pygame surface renderer class"""
    
    def __init__(self, surface: pg.Surface) -> None:

        self.surface = surface

    # PROPERTIES

    @property
    def width(self) -> int:
        return self.surface.get_width()

    @property
    def height(self) -> int:
        return self.surface.get_height()

    @property
    def dimension(self) -> Vec:
        return Vec(self.surface.get_width(), self.surface.get_height())

    # ABSTRACT METHODS

    @abstractmethod
    def render(self) -> None:
        """Render the surface

        *Returns nothing*
        """

        pass



"""Contains the main class of the application"""

# Standard modules
from abc import ABC, abstractmethod
import logging
import os
import sys

# Local modules
from .constants import *
from .utils import time
from .utils import log


class App(ABC):
    """Main class of the application"""

    _initialized = False

    def __init__(self, args: list[str] = None, *, name: str = "Sample Project", author: str = "Nikocraft",
                 version: str = "0.0.1", short_description: str = "A short description",
                 description: str = "This is a sample project for the nikocraft library ...",
                 details: str = "Some more details ...", log_path: str = "./logs",
                 log_thread: bool = False, log_date: bool = True) -> None:

        # General information
        self.args: list[str] = args if args else sys.argv
        self.debug: bool = "-d" in self.args or "--debug" in self.args
        self.name: str = name
        self.author: str = author
        self.version: str = version
        self.short_description: str = short_description
        self.description: str = description
        self.details: str = details
        self.log_path: str = log_path
        self.log_file: str = f"{self.log_path}/log_{time.datetime_f_ymd_hms()}.log"
        self.log_thread: bool = log_thread
        self.log_date: bool = log_date

        # Exit information
        self.exit_code: int = 0

        # Setup logging
        log.setup_log_directory(self.log_path)
        self.head: str = f"{self.name.upper()}\n{'-' * len(self.name)}\n\n{self.short_description}\n\nDESCRIPTION\n{self.description}\n\n" \
                         f"DETAILS\nVersion: {self.version}\nAuthor: {self.author}\nFramework: Nikocraft (v{VERSION})\n{self.details}\n"
        log.setup_log_file(self.log_file, self.head)
        self.logger: logging.Logger = log.create_logger(self.log_file, self.name.upper(), logging.DEBUG if self.debug else logging.INFO, log_thread=log_thread, log_date=log_date)

        # Log runtime information
        self.logger.info(f"Debug Mode: {'On' if self.debug else 'Off'}")
        self.logger.info(f"Run Path: {os.path.abspath('.')}")

        # Set initialized flag
        self._initialized: bool = True

    # METHODS

    def start(self) -> int:
        """Start the application

        Returns the exit code
        """

        # Check for initialization
        assert self._initialized, "Application was not initialized!"

        # Run application
        self.run()

        # Shutdown tasks
        self.quit()
        return self.exit_code

    # ABSTRACT METHODS

    @abstractmethod
    def run(self) -> None:
        """Application execution

        Called on starting -
        Don't call this method manually
        """

        pass

    def quit(self) -> None:
        """Shutdown tasks

        Called before exiting -
        Don't call this method manually
        """

        pass



"""Nikocraft Python Library"""

# Environment variables
import os
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = ""
del os

# Modules
from .utils import time
from .utils import file
from .utils import log

# Classes
from .app import App
from .utils.config import Config
from .utils.enum import Enum
from .window.window import Window
from .window.vector2d import Vec
from .window.vector3d import Vec3
from .window.rgb import RGB
from .window.rgb import RGBColor
from .window.clock import Clock
from .window.renderer import Renderer
from .window.debug_screen import DebugScreen
from .window.event_hook import EventHook

# Constants
from .constants import AUTHOR, VERSION, SCREEN_UPDATE_EVENT

__all__ = [
    "time",
    "file",
    "log",
    "App",
    "Config",
    "Enum",
    "Window",
    "Vec",
    "Vec3",
    "RGB",
    "RGBColor",
    "Clock",
    "Renderer",
    "DebugScreen",
    "EventHook",
    "AUTHOR",
    "VERSION",
    "SCREEN_UPDATE_EVENT"
]



# Local modules
from ..constants import SCREEN_UPDATE_EVENT
from .window import Window
from .renderer import Renderer
from .rgb import RGB


class DebugScreen(Renderer):
    """Renderer class for a debug screen"""
    
    def __init__(self, window: Window) -> None:

        super(DebugScreen, self).__init__(window.screen)

        self.window = window

        self.window.add_event_hook(f"DebugScreenSurfaceUpdate-{id(self)}", SCREEN_UPDATE_EVENT, lambda *args: setattr(self, "surface", self.window.screen))

    def render(self) -> None:

        font = self.window.font.get("calibri", 25, True)

        self.surface.blit(font.render(f"{self.window.app.name}   v{self.window.app.version}   (by {self.window.app.author})", True, RGB.WHITE), (0, 0))
        self.surface.blit(font.render(f"FPS: {self.window.clock.available_fps:.3f} ({self.window.clock.real_fps:.3f})", True, RGB.WHITE), (0, 35))
        self.surface.blit(font.render(f"Delta Time: {self.window.clock.delta_time_raw:.3f} ms ({self.window.clock.delta_time:.3f})", True, RGB.WHITE), (0, 60))
        self.surface.blit(font.render(f"Timings: {self.window.clock.frame_durations[-1]:.5f} ms", True, RGB.WHITE), (0, 85))
        self.surface.blit(font.render(f"   Event: {self.window.stat_event_time:.5f} ms", True, RGB.WHITE), (0, 110))
        self.surface.blit(font.render(f"   Render: {self.window.stat_render_time:.5f} ms", True, RGB.WHITE), (0, 135))
        self.surface.blit(font.render(f"   Other: {self.window.stat_other_time:.5f} ms", True, RGB.WHITE), (0, 160))
        self.surface.blit(font.render(f"Screen: {self.window.width} x {self.window.height} px", True, RGB.WHITE), (0, 195))




"""Contains the config class for saving JSON configurations"""

# Standard modules
from typing import Any
from abc import ABC
from logging import Logger

# Local modules
from ..utils import file


class Config(ABC):
    """Config class for saving JSON configurations"""

    def __init__(self, path: str, logger: Logger = None) -> None:

        self.path: str = path
        self.logger: Logger = logger

    # METHODS

    def load(self) -> None:
        """Load data from JSON file at path"""

        # Load file
        if self.logger:
            self.logger.debug(f"Load configuration file at '{self.path}' ...")
        data = file.load_json(self.path, self.logger, create=False)

        # Decode data
        if self.logger:
            self.logger.debug("Decode configuration attributes ...")
        for attribute in vars(self):
            if attribute in ("path", "logger"):
                continue
            result = self.load_attribute(attribute, data)
            if result is None:
                if attribute in data:
                    setattr(self, attribute, data[attribute])
            else:
                setattr(self, attribute, result)

    def save(self) -> None:
        """Save data to JSON file at path"""

        # Encode data
        data = {}
        if self.logger:
            self.logger.debug("Encode configuration attributes ...")
        for attribute in vars(self):
            if attribute in ("path", "logger"):
                continue
            result = self.save_attribute(attribute, getattr(self, attribute), data)
            if result is None:
                data[attribute] = getattr(self, attribute)

        # Save file
        if self.logger:
            self.logger.debug(f"Save configuration file at '{self.path}' ...")
        file.save_json(self.path, data, self.logger)

    # ABSTRACT METHODS

    def load_attribute(self, attribute: str, data: dict) -> Any | None:
        """Load an attribute from data (Return result or None for automatic loading)

        *Called on loading -
        Don't call this method manually*
        """

        pass

    def save_attribute(self, attribute: str, value: Any, data: dict) -> Any | None:
        """Save an attribute to data (Modify data or return None for automatic saving)

        *Called on saving -
        Don't call this method manually*
        """

        pass

    def load_extras(self, data: dict) -> None:
        """Additional loading tasks

        *Called on loading -
        Don't call this method manually*
        """

        pass

    def save_extras(self, data: dict) -> None:
        """Additional saving tasks

        *Called on saving -
        Don't call this method manually*
        """

        pass

    # OVERLOADS

    def __repr__(self) -> str:
        return f"Config[id={id(self)}, path={self.path}]"



"""Contains the RGB color class and common colors

Color source: https://www.webucator.com/article/python-color-constants-module/
"""

# Standard modules
from typing import Union, Self

# Local modules
from ..utils.enum import Enum


class RGBColor(tuple):
    """RGB color class"""

    def __new__(cls, red: int, green: int, blue: int) -> Self:
        return tuple.__new__(cls, (round(red), round(green), round(blue)))

    # PROPERTIES

    @property
    def red(self) -> int:
        return self[0]

    @property
    def green(self) -> int:
        return self[1]

    @property
    def blue(self) -> int:
        return self[2]

    @property
    def hex(self) -> str:
        return '#%02x%02x%02x' % self

    # CLASS METHODS

    @classmethod
    def from_hex(cls, value: str) -> Self:
        """Convert a hex color string to RGB"""
        value = value.replace("#", "")
        return RGBColor(*(int(value[i:i+2], 16) for i in (0, 2, 4)))

    # METHODS

    def rgba(self, value: int) -> tuple[int, int, int, int]:
        """Get a RGBA color tuple"""
        return self[0], self[1], self[2], value

    def list(self) -> list[int]:
        """Create a list with the red, green and blue values as elements"""
        return [*self]

    def dict(self, keys: tuple[str, str, str] = ("red", "green", "blue")) -> dict[str, int]:
        """Create a dictionary with the x and y values as value of specific keys"""
        return {keys[0]: self[0], keys[1]: self[1], keys[2]: self[2]}

    def clamp(self) -> Self:
        """Clamp the values to be greater equals 0 and lower equals 255"""
        clamped: list = self.list()
        for i in range(3):
            if clamped[i] < 0:
                clamped[i] = 0
            elif clamped[i] > 255:
                clamped[i] = 255
        return RGBColor(*clamped)

    def mix(self, color: Self) -> Self:
        """Mix two RGB colors"""
        return ((self + color) * 0.5).clamp()

    # OVERLOADS

    def __eq__(self, other: Self) -> bool:
        return self[0] == other[0] and self[1] == other[1] and self[2] == other[2]

    def __ne__(self, other: Self) -> bool:
        return self[0] != other[0] or self[1] != other[1] or self[2] != other[2]

    def __add__(self, other: Self) -> Self:
        return RGBColor(self[0] + other[0], self[1] + other[1], self[2] + other[2]).clamp()

    def __sub__(self, other: Self) -> Self:
        return RGBColor(self[0] - other[0], self[1] - other[1], self[2] - other[2]).clamp()

    def __mul__(self, factor: Union[int, float]) -> Self:
        return RGBColor(self[0] * factor, self[1] * factor, self[2] * factor).clamp()

    def __truediv__(self, divisor: Union[int, float]) -> Self:
        return RGBColor(self[0] / divisor, self[1] / divisor, self[2] / divisor).clamp()

    def __floordiv__(self, divisor: Union[int, float]) -> Self:
        return RGBColor(self[0] // divisor, self[1] // divisor, self[2] // divisor).clamp()

    def __pow__(self, power: float, modulo: float = None):
        return RGBColor(pow(self[0], power, modulo), pow(self[1], power, modulo), pow(self[2], power, modulo)).clamp()

    def __repr__(self) -> str:
        return f"RGB[id={id(self)}, red={self[0]}, green={self[1]}, blue={self[2]}]"

    def __str__(self) -> str:
        return f"({self[0]}, {self[1]}, {self[2]})"


class RGB(Enum):
    """RGB color library enum"""

    ALICEBLUE: RGBColor = RGBColor(240, 248, 255)
    ANTIQUEWHITE: RGBColor = RGBColor(250, 235, 215)
    ANTIQUEWHITE1: RGBColor = RGBColor(255, 239, 219)
    ANTIQUEWHITE2: RGBColor = RGBColor(238, 223, 204)
    ANTIQUEWHITE3: RGBColor = RGBColor(205, 192, 176)
    ANTIQUEWHITE4: RGBColor = RGBColor(139, 131, 120)
    AQUA: RGBColor = RGBColor(0, 255, 255)
    AQUAMARINE1: RGBColor = RGBColor(127, 255, 212)
    AQUAMARINE2: RGBColor = RGBColor(118, 238, 198)
    AQUAMARINE3: RGBColor = RGBColor(102, 205, 170)
    AQUAMARINE4: RGBColor = RGBColor(69, 139, 116)
    AZURE1: RGBColor = RGBColor(240, 255, 255)
    AZURE2: RGBColor = RGBColor(224, 238, 238)
    AZURE3: RGBColor = RGBColor(193, 205, 205)
    AZURE4: RGBColor = RGBColor(131, 139, 139)
    BANANA: RGBColor = RGBColor(227, 207, 87)
    BEIGE: RGBColor = RGBColor(245, 245, 220)
    BISQUE1: RGBColor = RGBColor(255, 228, 196)
    BISQUE2: RGBColor = RGBColor(238, 213, 183)
    BISQUE3: RGBColor = RGBColor(205, 183, 158)
    BISQUE4: RGBColor = RGBColor(139, 125, 107)
    BLACK: RGBColor = RGBColor(0, 0, 0)
    BLANCHEDALMOND: RGBColor = RGBColor(255, 235, 205)
    BLUE: RGBColor = RGBColor(0, 0, 255)
    BLUE2: RGBColor = RGBColor(0, 0, 238)
    BLUE3: RGBColor = RGBColor(0, 0, 205)
    BLUE4: RGBColor = RGBColor(0, 0, 139)
    BLUEVIOLET: RGBColor = RGBColor(138, 43, 226)
    BRICK: RGBColor = RGBColor(156, 102, 31)
    BROWN: RGBColor = RGBColor(165, 42, 42)
    BROWN1: RGBColor = RGBColor(255, 64, 64)
    BROWN2: RGBColor = RGBColor(238, 59, 59)
    BROWN3: RGBColor = RGBColor(205, 51, 51)
    BROWN4: RGBColor = RGBColor(139, 35, 35)
    BURLYWOOD: RGBColor = RGBColor(222, 184, 135)
    BURLYWOOD1: RGBColor = RGBColor(255, 211, 155)
    BURLYWOOD2: RGBColor = RGBColor(238, 197, 145)
    BURLYWOOD3: RGBColor = RGBColor(205, 170, 125)
    BURLYWOOD4: RGBColor = RGBColor(139, 115, 85)
    BURNTSIENNA: RGBColor = RGBColor(138, 54, 15)
    BURNTUMBER: RGBColor = RGBColor(138, 51, 36)
    CADETBLUE: RGBColor = RGBColor(95, 158, 160)
    CADETBLUE1: RGBColor = RGBColor(152, 245, 255)
    CADETBLUE2: RGBColor = RGBColor(142, 229, 238)
    CADETBLUE3: RGBColor = RGBColor(122, 197, 205)
    CADETBLUE4: RGBColor = RGBColor(83, 134, 139)
    CADMIUMORANGE: RGBColor = RGBColor(255, 97, 3)
    CADMIUMYELLOW: RGBColor = RGBColor(255, 153, 18)
    CARROT: RGBColor = RGBColor(237, 145, 33)
    CHARTREUSE1: RGBColor = RGBColor(127, 255, 0)
    CHARTREUSE2: RGBColor = RGBColor(118, 238, 0)
    CHARTREUSE3: RGBColor = RGBColor(102, 205, 0)
    CHARTREUSE4: RGBColor = RGBColor(69, 139, 0)
    CHOCOLATE: RGBColor = RGBColor(210, 105, 30)
    CHOCOLATE1: RGBColor = RGBColor(255, 127, 36)
    CHOCOLATE2: RGBColor = RGBColor(238, 118, 33)
    CHOCOLATE3: RGBColor = RGBColor(205, 102, 29)
    CHOCOLATE4: RGBColor = RGBColor(139, 69, 19)
    COBALT: RGBColor = RGBColor(61, 89, 171)
    COBALTGREEN: RGBColor = RGBColor(61, 145, 64)
    COLDGREY: RGBColor = RGBColor(128, 138, 135)
    CORAL: RGBColor = RGBColor(255, 127, 80)
    CORAL1: RGBColor = RGBColor(255, 114, 86)
    CORAL2: RGBColor = RGBColor(238, 106, 80)
    CORAL3: RGBColor = RGBColor(205, 91, 69)
    CORAL4: RGBColor = RGBColor(139, 62, 47)
    CORNFLOWERBLUE: RGBColor = RGBColor(100, 149, 237)
    CORNSILK1: RGBColor = RGBColor(255, 248, 220)
    CORNSILK2: RGBColor = RGBColor(238, 232, 205)
    CORNSILK3: RGBColor = RGBColor(205, 200, 177)
    CORNSILK4: RGBColor = RGBColor(139, 136, 120)
    CRIMSON: RGBColor = RGBColor(220, 20, 60)
    CYAN2: RGBColor = RGBColor(0, 238, 238)
    CYAN3: RGBColor = RGBColor(0, 205, 205)
    CYAN4: RGBColor = RGBColor(0, 139, 139)
    DARKGOLDENROD: RGBColor = RGBColor(184, 134, 11)
    DARKGOLDENROD1: RGBColor = RGBColor(255, 185, 15)
    DARKGOLDENROD2: RGBColor = RGBColor(238, 173, 14)
    DARKGOLDENROD3: RGBColor = RGBColor(205, 149, 12)
    DARKGOLDENROD4: RGBColor = RGBColor(139, 101, 8)
    DARKGRAY: RGBColor = RGBColor(169, 169, 169)
    DARKGREEN: RGBColor = RGBColor(0, 100, 0)
    DARKKHAKI: RGBColor = RGBColor(189, 183, 107)
    DARKOLIVEGREEN: RGBColor = RGBColor(85, 107, 47)
    DARKOLIVEGREEN1: RGBColor = RGBColor(202, 255, 112)
    DARKOLIVEGREEN2: RGBColor = RGBColor(188, 238, 104)
    DARKOLIVEGREEN3: RGBColor = RGBColor(162, 205, 90)
    DARKOLIVEGREEN4: RGBColor = RGBColor(110, 139, 61)
    DARKORANGE: RGBColor = RGBColor(255, 140, 0)
    DARKORANGE1: RGBColor = RGBColor(255, 127, 0)
    DARKORANGE2: RGBColor = RGBColor(238, 118, 0)
    DARKORANGE3: RGBColor = RGBColor(205, 102, 0)
    DARKORANGE4: RGBColor = RGBColor(139, 69, 0)
    DARKORCHID: RGBColor = RGBColor(153, 50, 204)
    DARKORCHID1: RGBColor = RGBColor(191, 62, 255)
    DARKORCHID2: RGBColor = RGBColor(178, 58, 238)
    DARKORCHID3: RGBColor = RGBColor(154, 50, 205)
    DARKORCHID4: RGBColor = RGBColor(104, 34, 139)
    DARKSALMON: RGBColor = RGBColor(233, 150, 122)
    DARKSEAGREEN: RGBColor = RGBColor(143, 188, 143)
    DARKSEAGREEN1: RGBColor = RGBColor(193, 255, 193)
    DARKSEAGREEN2: RGBColor = RGBColor(180, 238, 180)
    DARKSEAGREEN3: RGBColor = RGBColor(155, 205, 155)
    DARKSEAGREEN4: RGBColor = RGBColor(105, 139, 105)
    DARKSLATEBLUE: RGBColor = RGBColor(72, 61, 139)
    DARKSLATEGRAY: RGBColor = RGBColor(47, 79, 79)
    DARKSLATEGRAY1: RGBColor = RGBColor(151, 255, 255)
    DARKSLATEGRAY2: RGBColor = RGBColor(141, 238, 238)
    DARKSLATEGRAY3: RGBColor = RGBColor(121, 205, 205)
    DARKSLATEGRAY4: RGBColor = RGBColor(82, 139, 139)
    DARKTURQUOISE: RGBColor = RGBColor(0, 206, 209)
    DARKVIOLET: RGBColor = RGBColor(148, 0, 211)
    DEEPPINK1: RGBColor = RGBColor(255, 20, 147)
    DEEPPINK2: RGBColor = RGBColor(238, 18, 137)
    DEEPPINK3: RGBColor = RGBColor(205, 16, 118)
    DEEPPINK4: RGBColor = RGBColor(139, 10, 80)
    DEEPSKYBLUE1: RGBColor = RGBColor(0, 191, 255)
    DEEPSKYBLUE2: RGBColor = RGBColor(0, 178, 238)
    DEEPSKYBLUE3: RGBColor = RGBColor(0, 154, 205)
    DEEPSKYBLUE4: RGBColor = RGBColor(0, 104, 139)
    DIMGRAY: RGBColor = RGBColor(105, 105, 105)
    DODGERBLUE1: RGBColor = RGBColor(30, 144, 255)
    DODGERBLUE2: RGBColor = RGBColor(28, 134, 238)
    DODGERBLUE3: RGBColor = RGBColor(24, 116, 205)
    DODGERBLUE4: RGBColor = RGBColor(16, 78, 139)
    EGGSHELL: RGBColor = RGBColor(252, 230, 201)
    EMERALDGREEN: RGBColor = RGBColor(0, 201, 87)
    FIREBRICK: RGBColor = RGBColor(178, 34, 34)
    FIREBRICK1: RGBColor = RGBColor(255, 48, 48)
    FIREBRICK2: RGBColor = RGBColor(238, 44, 44)
    FIREBRICK3: RGBColor = RGBColor(205, 38, 38)
    FIREBRICK4: RGBColor = RGBColor(139, 26, 26)
    FLESH: RGBColor = RGBColor(255, 125, 64)
    FLORALWHITE: RGBColor = RGBColor(255, 250, 240)
    FORESTGREEN: RGBColor = RGBColor(34, 139, 34)
    GAINSBORO: RGBColor = RGBColor(220, 220, 220)
    GHOSTWHITE: RGBColor = RGBColor(248, 248, 255)
    GOLD1: RGBColor = RGBColor(255, 215, 0)
    GOLD2: RGBColor = RGBColor(238, 201, 0)
    GOLD3: RGBColor = RGBColor(205, 173, 0)
    GOLD4: RGBColor = RGBColor(139, 117, 0)
    GOLDENROD: RGBColor = RGBColor(218, 165, 32)
    GOLDENROD1: RGBColor = RGBColor(255, 193, 37)
    GOLDENROD2: RGBColor = RGBColor(238, 180, 34)
    GOLDENROD3: RGBColor = RGBColor(205, 155, 29)
    GOLDENROD4: RGBColor = RGBColor(139, 105, 20)
    GRAY: RGBColor = RGBColor(128, 128, 128)
    GRAY1: RGBColor = RGBColor(3, 3, 3)
    GRAY10: RGBColor = RGBColor(26, 26, 26)
    GRAY11: RGBColor = RGBColor(28, 28, 28)
    GRAY12: RGBColor = RGBColor(31, 31, 31)
    GRAY13: RGBColor = RGBColor(33, 33, 33)
    GRAY14: RGBColor = RGBColor(36, 36, 36)
    GRAY15: RGBColor = RGBColor(38, 38, 38)
    GRAY16: RGBColor = RGBColor(41, 41, 41)
    GRAY17: RGBColor = RGBColor(43, 43, 43)
    GRAY18: RGBColor = RGBColor(46, 46, 46)
    GRAY19: RGBColor = RGBColor(48, 48, 48)
    GRAY2: RGBColor = RGBColor(5, 5, 5)
    GRAY20: RGBColor = RGBColor(51, 51, 51)
    GRAY21: RGBColor = RGBColor(54, 54, 54)
    GRAY22: RGBColor = RGBColor(56, 56, 56)
    GRAY23: RGBColor = RGBColor(59, 59, 59)
    GRAY24: RGBColor = RGBColor(61, 61, 61)
    GRAY25: RGBColor = RGBColor(64, 64, 64)
    GRAY26: RGBColor = RGBColor(66, 66, 66)
    GRAY27: RGBColor = RGBColor(69, 69, 69)
    GRAY28: RGBColor = RGBColor(71, 71, 71)
    GRAY29: RGBColor = RGBColor(74, 74, 74)
    GRAY3: RGBColor = RGBColor(8, 8, 8)
    GRAY30: RGBColor = RGBColor(77, 77, 77)
    GRAY31: RGBColor = RGBColor(79, 79, 79)
    GRAY32: RGBColor = RGBColor(82, 82, 82)
    GRAY33: RGBColor = RGBColor(84, 84, 84)
    GRAY34: RGBColor = RGBColor(87, 87, 87)
    GRAY35: RGBColor = RGBColor(89, 89, 89)
    GRAY36: RGBColor = RGBColor(92, 92, 92)
    GRAY37: RGBColor = RGBColor(94, 94, 94)
    GRAY38: RGBColor = RGBColor(97, 97, 97)
    GRAY39: RGBColor = RGBColor(99, 99, 99)
    GRAY4: RGBColor = RGBColor(10, 10, 10)
    GRAY40: RGBColor = RGBColor(102, 102, 102)
    GRAY42: RGBColor = RGBColor(107, 107, 107)
    GRAY43: RGBColor = RGBColor(110, 110, 110)
    GRAY44: RGBColor = RGBColor(112, 112, 112)
    GRAY45: RGBColor = RGBColor(115, 115, 115)
    GRAY46: RGBColor = RGBColor(117, 117, 117)
    GRAY47: RGBColor = RGBColor(120, 120, 120)
    GRAY48: RGBColor = RGBColor(122, 122, 122)
    GRAY49: RGBColor = RGBColor(125, 125, 125)
    GRAY5: RGBColor = RGBColor(13, 13, 13)
    GRAY50: RGBColor = RGBColor(127, 127, 127)
    GRAY51: RGBColor = RGBColor(130, 130, 130)
    GRAY52: RGBColor = RGBColor(133, 133, 133)
    GRAY53: RGBColor = RGBColor(135, 135, 135)
    GRAY54: RGBColor = RGBColor(138, 138, 138)
    GRAY55: RGBColor = RGBColor(140, 140, 140)
    GRAY56: RGBColor = RGBColor(143, 143, 143)
    GRAY57: RGBColor = RGBColor(145, 145, 145)
    GRAY58: RGBColor = RGBColor(148, 148, 148)
    GRAY59: RGBColor = RGBColor(150, 150, 150)
    GRAY6: RGBColor = RGBColor(15, 15, 15)
    GRAY60: RGBColor = RGBColor(153, 153, 153)
    GRAY61: RGBColor = RGBColor(156, 156, 156)
    GRAY62: RGBColor = RGBColor(158, 158, 158)
    GRAY63: RGBColor = RGBColor(161, 161, 161)
    GRAY64: RGBColor = RGBColor(163, 163, 163)
    GRAY65: RGBColor = RGBColor(166, 166, 166)
    GRAY66: RGBColor = RGBColor(168, 168, 168)
    GRAY67: RGBColor = RGBColor(171, 171, 171)
    GRAY68: RGBColor = RGBColor(173, 173, 173)
    GRAY69: RGBColor = RGBColor(176, 176, 176)
    GRAY7: RGBColor = RGBColor(18, 18, 18)
    GRAY70: RGBColor = RGBColor(179, 179, 179)
    GRAY71: RGBColor = RGBColor(181, 181, 181)
    GRAY72: RGBColor = RGBColor(184, 184, 184)
    GRAY73: RGBColor = RGBColor(186, 186, 186)
    GRAY74: RGBColor = RGBColor(189, 189, 189)
    GRAY75: RGBColor = RGBColor(191, 191, 191)
    GRAY76: RGBColor = RGBColor(194, 194, 194)
    GRAY77: RGBColor = RGBColor(196, 196, 196)
    GRAY78: RGBColor = RGBColor(199, 199, 199)
    GRAY79: RGBColor = RGBColor(201, 201, 201)
    GRAY8: RGBColor = RGBColor(20, 20, 20)
    GRAY80: RGBColor = RGBColor(204, 204, 204)
    GRAY81: RGBColor = RGBColor(207, 207, 207)
    GRAY82: RGBColor = RGBColor(209, 209, 209)
    GRAY83: RGBColor = RGBColor(212, 212, 212)
    GRAY84: RGBColor = RGBColor(214, 214, 214)
    GRAY85: RGBColor = RGBColor(217, 217, 217)
    GRAY86: RGBColor = RGBColor(219, 219, 219)
    GRAY87: RGBColor = RGBColor(222, 222, 222)
    GRAY88: RGBColor = RGBColor(224, 224, 224)
    GRAY89: RGBColor = RGBColor(227, 227, 227)
    GRAY9: RGBColor = RGBColor(23, 23, 23)
    GRAY90: RGBColor = RGBColor(229, 229, 229)
    GRAY91: RGBColor = RGBColor(232, 232, 232)
    GRAY92: RGBColor = RGBColor(235, 235, 235)
    GRAY93: RGBColor = RGBColor(237, 237, 237)
    GRAY94: RGBColor = RGBColor(240, 240, 240)
    GRAY95: RGBColor = RGBColor(242, 242, 242)
    GRAY97: RGBColor = RGBColor(247, 247, 247)
    GRAY98: RGBColor = RGBColor(250, 250, 250)
    GRAY99: RGBColor = RGBColor(252, 252, 252)
    GREEN: RGBColor = RGBColor(0, 128, 0)
    GREEN1: RGBColor = RGBColor(0, 255, 0)
    GREEN2: RGBColor = RGBColor(0, 238, 0)
    GREEN3: RGBColor = RGBColor(0, 205, 0)
    GREEN4: RGBColor = RGBColor(0, 139, 0)
    GREENYELLOW: RGBColor = RGBColor(173, 255, 47)
    HONEYDEW1: RGBColor = RGBColor(240, 255, 240)
    HONEYDEW2: RGBColor = RGBColor(224, 238, 224)
    HONEYDEW3: RGBColor = RGBColor(193, 205, 193)
    HONEYDEW4: RGBColor = RGBColor(131, 139, 131)
    HOTPINK: RGBColor = RGBColor(255, 105, 180)
    HOTPINK1: RGBColor = RGBColor(255, 110, 180)
    HOTPINK2: RGBColor = RGBColor(238, 106, 167)
    HOTPINK3: RGBColor = RGBColor(205, 96, 144)
    HOTPINK4: RGBColor = RGBColor(139, 58, 98)
    INDIANRED: RGBColor = RGBColor(205, 92, 92)
    INDIANRED1: RGBColor = RGBColor(255, 106, 106)
    INDIANRED2: RGBColor = RGBColor(238, 99, 99)
    INDIANRED3: RGBColor = RGBColor(205, 85, 85)
    INDIANRED4: RGBColor = RGBColor(139, 58, 58)
    INDIGO: RGBColor = RGBColor(75, 0, 130)
    IVORY1: RGBColor = RGBColor(255, 255, 240)
    IVORY2: RGBColor = RGBColor(238, 238, 224)
    IVORY3: RGBColor = RGBColor(205, 205, 193)
    IVORY4: RGBColor = RGBColor(139, 139, 131)
    IVORYBLACK: RGBColor = RGBColor(41, 36, 33)
    KHAKI: RGBColor = RGBColor(240, 230, 140)
    KHAKI1: RGBColor = RGBColor(255, 246, 143)
    KHAKI2: RGBColor = RGBColor(238, 230, 133)
    KHAKI3: RGBColor = RGBColor(205, 198, 115)
    KHAKI4: RGBColor = RGBColor(139, 134, 78)
    LAVENDER: RGBColor = RGBColor(230, 230, 250)
    LAVENDERBLUSH1: RGBColor = RGBColor(255, 240, 245)
    LAVENDERBLUSH2: RGBColor = RGBColor(238, 224, 229)
    LAVENDERBLUSH3: RGBColor = RGBColor(205, 193, 197)
    LAVENDERBLUSH4: RGBColor = RGBColor(139, 131, 134)
    LAWNGREEN: RGBColor = RGBColor(124, 252, 0)
    LEMONCHIFFON1: RGBColor = RGBColor(255, 250, 205)
    LEMONCHIFFON2: RGBColor = RGBColor(238, 233, 191)
    LEMONCHIFFON3: RGBColor = RGBColor(205, 201, 165)
    LEMONCHIFFON4: RGBColor = RGBColor(139, 137, 112)
    LIGHTBLUE: RGBColor = RGBColor(173, 216, 230)
    LIGHTBLUE1: RGBColor = RGBColor(191, 239, 255)
    LIGHTBLUE2: RGBColor = RGBColor(178, 223, 238)
    LIGHTBLUE3: RGBColor = RGBColor(154, 192, 205)
    LIGHTBLUE4: RGBColor = RGBColor(104, 131, 139)
    LIGHTCORAL: RGBColor = RGBColor(240, 128, 128)
    LIGHTCYAN1: RGBColor = RGBColor(224, 255, 255)
    LIGHTCYAN2: RGBColor = RGBColor(209, 238, 238)
    LIGHTCYAN3: RGBColor = RGBColor(180, 205, 205)
    LIGHTCYAN4: RGBColor = RGBColor(122, 139, 139)
    LIGHTGOLDENROD1: RGBColor = RGBColor(255, 236, 139)
    LIGHTGOLDENROD2: RGBColor = RGBColor(238, 220, 130)
    LIGHTGOLDENROD3: RGBColor = RGBColor(205, 190, 112)
    LIGHTGOLDENROD4: RGBColor = RGBColor(139, 129, 76)
    LIGHTGOLDENRODYELLOW: RGBColor = RGBColor(250, 250, 210)
    LIGHTGREY: RGBColor = RGBColor(211, 211, 211)
    LIGHTPINK: RGBColor = RGBColor(255, 182, 193)
    LIGHTPINK1: RGBColor = RGBColor(255, 174, 185)
    LIGHTPINK2: RGBColor = RGBColor(238, 162, 173)
    LIGHTPINK3: RGBColor = RGBColor(205, 140, 149)
    LIGHTPINK4: RGBColor = RGBColor(139, 95, 101)
    LIGHTSALMON1: RGBColor = RGBColor(255, 160, 122)
    LIGHTSALMON2: RGBColor = RGBColor(238, 149, 114)
    LIGHTSALMON3: RGBColor = RGBColor(205, 129, 98)
    LIGHTSALMON4: RGBColor = RGBColor(139, 87, 66)
    LIGHTSEAGREEN: RGBColor = RGBColor(32, 178, 170)
    LIGHTSKYBLUE: RGBColor = RGBColor(135, 206, 250)
    LIGHTSKYBLUE1: RGBColor = RGBColor(176, 226, 255)
    LIGHTSKYBLUE2: RGBColor = RGBColor(164, 211, 238)
    LIGHTSKYBLUE3: RGBColor = RGBColor(141, 182, 205)
    LIGHTSKYBLUE4: RGBColor = RGBColor(96, 123, 139)
    LIGHTSLATEBLUE: RGBColor = RGBColor(132, 112, 255)
    LIGHTSLATEGRAY: RGBColor = RGBColor(119, 136, 153)
    LIGHTSTEELBLUE: RGBColor = RGBColor(176, 196, 222)
    LIGHTSTEELBLUE1: RGBColor = RGBColor(202, 225, 255)
    LIGHTSTEELBLUE2: RGBColor = RGBColor(188, 210, 238)
    LIGHTSTEELBLUE3: RGBColor = RGBColor(162, 181, 205)
    LIGHTSTEELBLUE4: RGBColor = RGBColor(110, 123, 139)
    LIGHTYELLOW1: RGBColor = RGBColor(255, 255, 224)
    LIGHTYELLOW2: RGBColor = RGBColor(238, 238, 209)
    LIGHTYELLOW3: RGBColor = RGBColor(205, 205, 180)
    LIGHTYELLOW4: RGBColor = RGBColor(139, 139, 122)
    LIMEGREEN: RGBColor = RGBColor(50, 205, 50)
    LINEN: RGBColor = RGBColor(250, 240, 230)
    MAGENTA: RGBColor = RGBColor(255, 0, 255)
    MAGENTA2: RGBColor = RGBColor(238, 0, 238)
    MAGENTA3: RGBColor = RGBColor(205, 0, 205)
    MAGENTA4: RGBColor = RGBColor(139, 0, 139)
    MANGANESEBLUE: RGBColor = RGBColor(3, 168, 158)
    MAROON: RGBColor = RGBColor(128, 0, 0)
    MAROON1: RGBColor = RGBColor(255, 52, 179)
    MAROON2: RGBColor = RGBColor(238, 48, 167)
    MAROON3: RGBColor = RGBColor(205, 41, 144)
    MAROON4: RGBColor = RGBColor(139, 28, 98)
    MEDIUMORCHID: RGBColor = RGBColor(186, 85, 211)
    MEDIUMORCHID1: RGBColor = RGBColor(224, 102, 255)
    MEDIUMORCHID2: RGBColor = RGBColor(209, 95, 238)
    MEDIUMORCHID3: RGBColor = RGBColor(180, 82, 205)
    MEDIUMORCHID4: RGBColor = RGBColor(122, 55, 139)
    MEDIUMPURPLE: RGBColor = RGBColor(147, 112, 219)
    MEDIUMPURPLE1: RGBColor = RGBColor(171, 130, 255)
    MEDIUMPURPLE2: RGBColor = RGBColor(159, 121, 238)
    MEDIUMPURPLE3: RGBColor = RGBColor(137, 104, 205)
    MEDIUMPURPLE4: RGBColor = RGBColor(93, 71, 139)
    MEDIUMSEAGREEN: RGBColor = RGBColor(60, 179, 113)
    MEDIUMSLATEBLUE: RGBColor = RGBColor(123, 104, 238)
    MEDIUMSPRINGGREEN: RGBColor = RGBColor(0, 250, 154)
    MEDIUMTURQUOISE: RGBColor = RGBColor(72, 209, 204)
    MEDIUMVIOLETRED: RGBColor = RGBColor(199, 21, 133)
    MELON: RGBColor = RGBColor(227, 168, 105)
    MIDNIGHTBLUE: RGBColor = RGBColor(25, 25, 112)
    MINT: RGBColor = RGBColor(189, 252, 201)
    MINTCREAM: RGBColor = RGBColor(245, 255, 250)
    MISTYROSE1: RGBColor = RGBColor(255, 228, 225)
    MISTYROSE2: RGBColor = RGBColor(238, 213, 210)
    MISTYROSE3: RGBColor = RGBColor(205, 183, 181)
    MISTYROSE4: RGBColor = RGBColor(139, 125, 123)
    MOCCASIN: RGBColor = RGBColor(255, 228, 181)
    NAVAJOWHITE1: RGBColor = RGBColor(255, 222, 173)
    NAVAJOWHITE2: RGBColor = RGBColor(238, 207, 161)
    NAVAJOWHITE3: RGBColor = RGBColor(205, 179, 139)
    NAVAJOWHITE4: RGBColor = RGBColor(139, 121, 94)
    NAVY: RGBColor = RGBColor(0, 0, 128)
    OLDLACE: RGBColor = RGBColor(253, 245, 230)
    OLIVE: RGBColor = RGBColor(128, 128, 0)
    OLIVEDRAB: RGBColor = RGBColor(107, 142, 35)
    OLIVEDRAB1: RGBColor = RGBColor(192, 255, 62)
    OLIVEDRAB2: RGBColor = RGBColor(179, 238, 58)
    OLIVEDRAB3: RGBColor = RGBColor(154, 205, 50)
    OLIVEDRAB4: RGBColor = RGBColor(105, 139, 34)
    ORANGE: RGBColor = RGBColor(255, 128, 0)
    ORANGE1: RGBColor = RGBColor(255, 165, 0)
    ORANGE2: RGBColor = RGBColor(238, 154, 0)
    ORANGE3: RGBColor = RGBColor(205, 133, 0)
    ORANGE4: RGBColor = RGBColor(139, 90, 0)
    ORANGERED1: RGBColor = RGBColor(255, 69, 0)
    ORANGERED2: RGBColor = RGBColor(238, 64, 0)
    ORANGERED3: RGBColor = RGBColor(205, 55, 0)
    ORANGERED4: RGBColor = RGBColor(139, 37, 0)
    ORCHID: RGBColor = RGBColor(218, 112, 214)
    ORCHID1: RGBColor = RGBColor(255, 131, 250)
    ORCHID2: RGBColor = RGBColor(238, 122, 233)
    ORCHID3: RGBColor = RGBColor(205, 105, 201)
    ORCHID4: RGBColor = RGBColor(139, 71, 137)
    PALEGOLDENROD: RGBColor = RGBColor(238, 232, 170)
    PALEGREEN: RGBColor = RGBColor(152, 251, 152)
    PALEGREEN1: RGBColor = RGBColor(154, 255, 154)
    PALEGREEN2: RGBColor = RGBColor(144, 238, 144)
    PALEGREEN3: RGBColor = RGBColor(124, 205, 124)
    PALEGREEN4: RGBColor = RGBColor(84, 139, 84)
    PALETURQUOISE1: RGBColor = RGBColor(187, 255, 255)
    PALETURQUOISE2: RGBColor = RGBColor(174, 238, 238)
    PALETURQUOISE3: RGBColor = RGBColor(150, 205, 205)
    PALETURQUOISE4: RGBColor = RGBColor(102, 139, 139)
    PALEVIOLETRED: RGBColor = RGBColor(219, 112, 147)
    PALEVIOLETRED1: RGBColor = RGBColor(255, 130, 171)
    PALEVIOLETRED2: RGBColor = RGBColor(238, 121, 159)
    PALEVIOLETRED3: RGBColor = RGBColor(205, 104, 137)
    PALEVIOLETRED4: RGBColor = RGBColor(139, 71, 93)
    PAPAYAWHIP: RGBColor = RGBColor(255, 239, 213)
    PEACHPUFF1: RGBColor = RGBColor(255, 218, 185)
    PEACHPUFF2: RGBColor = RGBColor(238, 203, 173)
    PEACHPUFF3: RGBColor = RGBColor(205, 175, 149)
    PEACHPUFF4: RGBColor = RGBColor(139, 119, 101)
    PEACOCK: RGBColor = RGBColor(51, 161, 201)
    PINK: RGBColor = RGBColor(255, 192, 203)
    PINK1: RGBColor = RGBColor(255, 181, 197)
    PINK2: RGBColor = RGBColor(238, 169, 184)
    PINK3: RGBColor = RGBColor(205, 145, 158)
    PINK4: RGBColor = RGBColor(139, 99, 108)
    PLUM: RGBColor = RGBColor(221, 160, 221)
    PLUM1: RGBColor = RGBColor(255, 187, 255)
    PLUM2: RGBColor = RGBColor(238, 174, 238)
    PLUM3: RGBColor = RGBColor(205, 150, 205)
    PLUM4: RGBColor = RGBColor(139, 102, 139)
    POWDERBLUE: RGBColor = RGBColor(176, 224, 230)
    PURPLE: RGBColor = RGBColor(128, 0, 128)
    PURPLE1: RGBColor = RGBColor(155, 48, 255)
    PURPLE2: RGBColor = RGBColor(145, 44, 238)
    PURPLE3: RGBColor = RGBColor(125, 38, 205)
    PURPLE4: RGBColor = RGBColor(85, 26, 139)
    RASPBERRY: RGBColor = RGBColor(135, 38, 87)
    RAWSIENNA: RGBColor = RGBColor(199, 97, 20)
    RED1: RGBColor = RGBColor(255, 0, 0)
    RED2: RGBColor = RGBColor(238, 0, 0)
    RED3: RGBColor = RGBColor(205, 0, 0)
    RED4: RGBColor = RGBColor(139, 0, 0)
    ROSYBROWN: RGBColor = RGBColor(188, 143, 143)
    ROSYBROWN1: RGBColor = RGBColor(255, 193, 193)
    ROSYBROWN2: RGBColor = RGBColor(238, 180, 180)
    ROSYBROWN3: RGBColor = RGBColor(205, 155, 155)
    ROSYBROWN4: RGBColor = RGBColor(139, 105, 105)
    ROYALBLUE: RGBColor = RGBColor(65, 105, 225)
    ROYALBLUE1: RGBColor = RGBColor(72, 118, 255)
    ROYALBLUE2: RGBColor = RGBColor(67, 110, 238)
    ROYALBLUE3: RGBColor = RGBColor(58, 95, 205)
    ROYALBLUE4: RGBColor = RGBColor(39, 64, 139)
    SALMON: RGBColor = RGBColor(250, 128, 114)
    SALMON1: RGBColor = RGBColor(255, 140, 105)
    SALMON2: RGBColor = RGBColor(238, 130, 98)
    SALMON3: RGBColor = RGBColor(205, 112, 84)
    SALMON4: RGBColor = RGBColor(139, 76, 57)
    SANDYBROWN: RGBColor = RGBColor(244, 164, 96)
    SAPGREEN: RGBColor = RGBColor(48, 128, 20)
    SEAGREEN1: RGBColor = RGBColor(84, 255, 159)
    SEAGREEN2: RGBColor = RGBColor(78, 238, 148)
    SEAGREEN3: RGBColor = RGBColor(67, 205, 128)
    SEAGREEN4: RGBColor = RGBColor(46, 139, 87)
    SEASHELL1: RGBColor = RGBColor(255, 245, 238)
    SEASHELL2: RGBColor = RGBColor(238, 229, 222)
    SEASHELL3: RGBColor = RGBColor(205, 197, 191)
    SEASHELL4: RGBColor = RGBColor(139, 134, 130)
    SEPIA: RGBColor = RGBColor(94, 38, 18)
    SGIBEET: RGBColor = RGBColor(142, 56, 142)
    SGIBRIGHTGRAY: RGBColor = RGBColor(197, 193, 170)
    SGICHARTREUSE: RGBColor = RGBColor(113, 198, 113)
    SGIDARKGRAY: RGBColor = RGBColor(85, 85, 85)
    SGIGRAY12: RGBColor = RGBColor(30, 30, 30)
    SGIGRAY16: RGBColor = RGBColor(40, 40, 40)
    SGIGRAY32: RGBColor = RGBColor(81, 81, 81)
    SGIGRAY36: RGBColor = RGBColor(91, 91, 91)
    SGIGRAY52: RGBColor = RGBColor(132, 132, 132)
    SGIGRAY56: RGBColor = RGBColor(142, 142, 142)
    SGIGRAY72: RGBColor = RGBColor(183, 183, 183)
    SGIGRAY76: RGBColor = RGBColor(193, 193, 193)
    SGIGRAY92: RGBColor = RGBColor(234, 234, 234)
    SGIGRAY96: RGBColor = RGBColor(244, 244, 244)
    SGILIGHTBLUE: RGBColor = RGBColor(125, 158, 192)
    SGILIGHTGRAY: RGBColor = RGBColor(170, 170, 170)
    SGIOLIVEDRAB: RGBColor = RGBColor(142, 142, 56)
    SGISALMON: RGBColor = RGBColor(198, 113, 113)
    SGISLATEBLUE: RGBColor = RGBColor(113, 113, 198)
    SGITEAL: RGBColor = RGBColor(56, 142, 142)
    SIENNA: RGBColor = RGBColor(160, 82, 45)
    SIENNA1: RGBColor = RGBColor(255, 130, 71)
    SIENNA2: RGBColor = RGBColor(238, 121, 66)
    SIENNA3: RGBColor = RGBColor(205, 104, 57)
    SIENNA4: RGBColor = RGBColor(139, 71, 38)
    SILVER: RGBColor = RGBColor(192, 192, 192)
    SKYBLUE: RGBColor = RGBColor(135, 206, 235)
    SKYBLUE1: RGBColor = RGBColor(135, 206, 255)
    SKYBLUE2: RGBColor = RGBColor(126, 192, 238)
    SKYBLUE3: RGBColor = RGBColor(108, 166, 205)
    SKYBLUE4: RGBColor = RGBColor(74, 112, 139)
    SLATEBLUE: RGBColor = RGBColor(106, 90, 205)
    SLATEBLUE1: RGBColor = RGBColor(131, 111, 255)
    SLATEBLUE2: RGBColor = RGBColor(122, 103, 238)
    SLATEBLUE3: RGBColor = RGBColor(105, 89, 205)
    SLATEBLUE4: RGBColor = RGBColor(71, 60, 139)
    SLATEGRAY: RGBColor = RGBColor(112, 128, 144)
    SLATEGRAY1: RGBColor = RGBColor(198, 226, 255)
    SLATEGRAY2: RGBColor = RGBColor(185, 211, 238)
    SLATEGRAY3: RGBColor = RGBColor(159, 182, 205)
    SLATEGRAY4: RGBColor = RGBColor(108, 123, 139)
    SNOW1: RGBColor = RGBColor(255, 250, 250)
    SNOW2: RGBColor = RGBColor(238, 233, 233)
    SNOW3: RGBColor = RGBColor(205, 201, 201)
    SNOW4: RGBColor = RGBColor(139, 137, 137)
    SPRINGGREEN: RGBColor = RGBColor(0, 255, 127)
    SPRINGGREEN1: RGBColor = RGBColor(0, 238, 118)
    SPRINGGREEN2: RGBColor = RGBColor(0, 205, 102)
    SPRINGGREEN3: RGBColor = RGBColor(0, 139, 69)
    STEELBLUE: RGBColor = RGBColor(70, 130, 180)
    STEELBLUE1: RGBColor = RGBColor(99, 184, 255)
    STEELBLUE2: RGBColor = RGBColor(92, 172, 238)
    STEELBLUE3: RGBColor = RGBColor(79, 148, 205)
    STEELBLUE4: RGBColor = RGBColor(54, 100, 139)
    TAN: RGBColor = RGBColor(210, 180, 140)
    TAN1: RGBColor = RGBColor(255, 165, 79)
    TAN2: RGBColor = RGBColor(238, 154, 73)
    TAN3: RGBColor = RGBColor(205, 133, 63)
    TAN4: RGBColor = RGBColor(139, 90, 43)
    TEAL: RGBColor = RGBColor(0, 128, 128)
    THISTLE: RGBColor = RGBColor(216, 191, 216)
    THISTLE1: RGBColor = RGBColor(255, 225, 255)
    THISTLE2: RGBColor = RGBColor(238, 210, 238)
    THISTLE3: RGBColor = RGBColor(205, 181, 205)
    THISTLE4: RGBColor = RGBColor(139, 123, 139)
    TOMATO1: RGBColor = RGBColor(255, 99, 71)
    TOMATO2: RGBColor = RGBColor(238, 92, 66)
    TOMATO3: RGBColor = RGBColor(205, 79, 57)
    TOMATO4: RGBColor = RGBColor(139, 54, 38)
    TURQUOISE: RGBColor = RGBColor(64, 224, 208)
    TURQUOISE1: RGBColor = RGBColor(0, 245, 255)
    TURQUOISE2: RGBColor = RGBColor(0, 229, 238)
    TURQUOISE3: RGBColor = RGBColor(0, 197, 205)
    TURQUOISE4: RGBColor = RGBColor(0, 134, 139)
    TURQUOISEBLUE: RGBColor = RGBColor(0, 199, 140)
    VIOLET: RGBColor = RGBColor(238, 130, 238)
    VIOLETRED: RGBColor = RGBColor(208, 32, 144)
    VIOLETRED1: RGBColor = RGBColor(255, 62, 150)
    VIOLETRED2: RGBColor = RGBColor(238, 58, 140)
    VIOLETRED3: RGBColor = RGBColor(205, 50, 120)
    VIOLETRED4: RGBColor = RGBColor(139, 34, 82)
    WARMGREY: RGBColor = RGBColor(128, 128, 105)
    WHEAT: RGBColor = RGBColor(245, 222, 179)
    WHEAT1: RGBColor = RGBColor(255, 231, 186)
    WHEAT2: RGBColor = RGBColor(238, 216, 174)
    WHEAT3: RGBColor = RGBColor(205, 186, 150)
    WHEAT4: RGBColor = RGBColor(139, 126, 102)
    WHITE: RGBColor = RGBColor(255, 255, 255)
    WHITESMOKE: RGBColor = RGBColor(245, 245, 245)
    YELLOW1: RGBColor = RGBColor(255, 255, 0)
    YELLOW2: RGBColor = RGBColor(238, 238, 0)
    YELLOW3: RGBColor = RGBColor(205, 205, 0)
    YELLOW4: RGBColor = RGBColor(139, 139, 0)



"""Contains the window class for GUI management"""

# Standard modules
from typing import Callable, Self
import os
import ctypes
import logging

# External modules
os.environ["PYGAME_HIDE_SUPPORT_PROMPT"] = ""
import pygame as pg

# Local modules
from ..constants import *
from ..app import App
from ..utils import time
from .vector2d import Vec
from .clock import Clock
from .font import FontManager
from .event_hook import EventHook


class Window:
    """Window class for the GUI management"""

    _initialized = False

    def __init__(self, app: App) -> None:

        # App
        self.app = app

        # General information
        self.screen: pg.Surface = pg.Surface((0, 0))
        self.target_dimension: Vec = Vec(DEFAULT_WIDTH, DEFAULT_HEIGHT)
        self.running: bool = False
        self.clock: Clock = Clock(DEFAULT_FPS)
        self.flags: int = 0
        self.option_auto_update_screen: bool = True
        self.option_auto_quit: bool = True

        # Statistics
        self.stat_event_time: float = 0
        self.stat_render_time: float = 0
        self.stat_other_time: float = 0

        # Initialize pygame
        pg.init()

        # Initialize font manager
        self.font: FontManager = FontManager()

        # Event hooks
        self.event_hooks: list[EventHook] = []

        # Set initialized flag
        self._initialized = True

    # PROPERTIES

    @property
    def logger(self) -> logging.Logger:
        return self.app.logger

    @property
    def width(self) -> int:
        return self.screen.get_width()

    @property
    def height(self) -> int:
        return self.screen.get_height()

    @property
    def dimension(self) -> Vec:
        return Vec(self.screen.get_width(), self.screen.get_height())

    # METHODS

    def open(self) -> None:
        """Open the window

        *Returns nothing*
        """

        # Check for initialization
        assert self._initialized, "Window was not initialized!"

        # Open window
        self.screen: pg.Surface = pg.display.set_mode(self.target_dimension, self.flags)
        pg.event.post(pg.event.Event(SCREEN_UPDATE_EVENT))
        self.init()

        # Window loop
        self.running = True
        while self.running:

            # Clock tick
            self.clock.tick()

            # Event handling
            with time.benchmark(lambda result: setattr(self, "stat_event_time", result)):
                for event in pg.event.get():

                    # Window event handler
                    self.event(event)

                    # Event hooks
                    for hook in self.event_hooks:
                        for event_type in hook.events:
                            if event.type == event_type:
                                for handler in hook.handlers:
                                    handler(event, self, hook.data)
                                break

                    # Default event handlers
                    if event.type == pg.QUIT and self.option_auto_quit:
                        self.running = False

            # Screen rendering
            with time.benchmark(lambda result: setattr(self, "stat_render_time", result)):
                self.render()
                if self.option_auto_update_screen:
                    pg.display.flip()

            # Statistics
            self.stat_other_time: float = self.clock.frame_durations[-1] - self.stat_event_time - self.stat_render_time
            if self.stat_other_time < 0:
                self.stat_other_time = 0

        # Shutdown
        self.quit()
        pg.quit()

    def add_event_hook(self, name: str, event: int | tuple[int, ...], handler: tuple[Callable[[pg.event.Event, Self, dict], None], ...] |
                       Callable[[pg.event.Event, Self, dict], None], data: dict = None) -> EventHook:
        """Add a new event hook"""
        hook = EventHook(name, event if isinstance(event, tuple) else (event,),
                         handler if isinstance(handler, tuple) else (handler,), data if data is not None else {})
        self.event_hooks.append(hook)
        return hook

    def remove_event_hook(self, name: str) -> bool:
        """Remove a event hook"""
        for hook in self.event_hooks:
            if hook.name == name:
                self.event_hooks.remove(hook)
                return True
        return False

    # ABSTRACT METHODS

    def init(self) -> None:
        """Startup tasks

        *Called after window is opened -
        Don't call this method manually*
        """

        pass

    def event(self, event: pg.event.Event) -> None:
        """Handle pygame event

        *Called when event is fired*
        """

        pass

    def render(self) -> None:
        """Render screen

        *Called every frame*
        """

        pass

    def quit(self) -> None:
        """Shutdown tasks

        *Called before closing -
        Don't call this method manually*
        """

        pass

    # STATIC METHODS

    @staticmethod
    def disable_resolution_scaling() -> None:
        """Disable resolution scaling on Windows

        Returns nothing
        """

        if os.name == "nt":
            ctypes.windll.user32.SetProcessDPIAware()
